from __future__ import absolute_import
from __future__ import print_function
import collections
import sys
from xml.parsers import expat
from . import Extractor
from . import Message


def _open(filename):
    """Injection point for tests."""
    return open(filename, 'rb')


class ZCMLExtractor(Extractor):
    '''Zope Configuration Markup Language (ZCML)'''
    extensions = ['.zcml']
    ATTRIBUTES = set(['title', 'description'])

    def __call__(self, filename, options, fileobj=None, lineno=0):
        self.filename = filename
        self.target_domain = options.domain
        self.messages = []
        self.parser = expat.ParserCreate()
        self.parser.StartElementHandler = self.StartElementHandler
        self.parser.EndElementHandler = self.EndElementHandler
        self.domainstack = collections.deque()
        if fileobj is None:
            fileobj = _open(filename)
        try:
            self.parser.ParseFile(fileobj)
        except expat.ExpatError as e:
            print('Aborting due to parse error in %s: %s' %
                            (filename, e), file=sys.stderr)
            sys.exit(1)
        return self.messages

    def add_message(self, msgid):
        self.messages.append(
                Message(None, msgid, None, [], u'', u'',
                    (self.filename, (self.parser.CurrentLineNumber))))

    def StartElementHandler(self, name, attributes):
        if 'i18n_domain' in attributes:
            self.domainstack.append(attributes["i18n_domain"])
        elif self.domainstack:
            self.domainstack.append(self.domainstack[-1])

        if not self.domainstack:
            return

        if self.target_domain in [None, self.domainstack[-1]]:
            for (key, value) in attributes.items():
                if key in self.ATTRIBUTES:
                    self.add_message(value)

    def EndElementHandler(self, name):
        if self.domainstack:
            self.domainstack.pop()



from __future__ import absolute_import
from __future__ import print_function
import ast
import collections
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
import re
import sys
from chameleon.namespaces import I18N_NS
from chameleon.namespaces import TAL_NS
from chameleon.program import ElementProgram
from chameleon.zpt.program import MacroProgram
from chameleon.tal import parse_defines
from chameleon.tales import split_parts
from chameleon.utils import decode_htmlentities

from .python import _extract_python
from . import Extractor
from . import Message


def _open(filename):
    """Injection point for tests."""
    return open(filename, 'rb')


ENGINE_PREFIX = re.compile(r'^\s*([a-z\-_]+):\s*')
STRUCTURE_PREFIX = re.compile(r'\s*(structure|text)\s+(.*)', re.DOTALL)
WHITESPACE = re.compile(u"\s+")
EXPRESSION = re.compile(u"\s*\${(.*?)}\s*")
UNDERSCORE_CALL = re.compile("_\(.*\)")


class TranslateContext(object):
    def __init__(self, domain, msgctxt, msgid, comment, filename, lineno):
        self.domain = domain
        self.msgctxt = msgctxt
        self.msgid = msgid
        self.text = []
        self.filename = filename
        self.lineno = lineno
        self.parent = None
        self.children = OrderedDict()
        self.comment = comment

    def add_text(self, text):
        self.text.append(text)

    def add_element(self, element):
        attributes = element['ns_attrs']
        name = attributes.get((I18N_NS, 'name'))
        if name:
            self.text.append(u'${%s}' % name)
        else:
            self.text.append(u'<dynamic element>')

    def register_child(self, element, context):
        attributes = element['ns_attrs']
        name = attributes.get((I18N_NS, 'name'))
        if name:
            self.children[name] = context

    def ignore(self):
        text = u''.join(self.text).strip()
        text = WHITESPACE.sub(u' ', text)
        text = EXPRESSION.sub(u'', text)
        return not text

    def full_text(self):
        text = u''.join(self.text).strip()
        text = WHITESPACE.sub(u' ', text)
        return text

    def message(self):
        text = self.full_text()
        if not self.msgid:
            self.msgid = text
            text = u''
        comments = []
        if self.comment:
            comments.append(self.comment)
        if text:
            comments.append(u'Default: %s' % text)
        for (name, context) in self.children.items():
            comments.append(u'Canonical text for ${%s} is: "%s"' %
                    (name, context.full_text()))
        if self.parent:
            comments.append(u'Used in sentence: "%s"' %
                    self.parent.full_text())
        return Message(self.msgctxt, self.msgid, None, [],
                u'\n'.join(comments), u'',
                (self.filename, self.lineno))


def get_newline_count(s):
    s = s or ''
    return len(s.split('\n')) - 1


def get_plain_attrs(attrs):
    plain_attrs = dict()
    offset = 0
    for attr in attrs:
        offset += get_newline_count(attr['space'] + attr['name'] + attr['eq'] + attr['quote'])
        post_offset = offset + get_newline_count(attr['value'] + attr['quote'])
        plain_attrs[attr['name'].split(':')[-1]] = (attr['value'], offset, post_offset)
        offset = post_offset
    return plain_attrs


class ChameleonExtractor(Extractor, ElementProgram):
    '''Chameleon templates (defaults to Python expressions)'''
    extensions = ['.pt']
    DEFAULT_NAMESPACES = MacroProgram.DEFAULT_NAMESPACES
    default_config = {
            'default-engine': 'python',
            }

    def __call__(self, filename, options, fileobj=None, lineno=0):
        self.options = options
        self.filename = filename
        self.target_domain = options.domain
        self.messages = []
        self.domainstack = collections.deque([(None, None)])
        self.translatestack = collections.deque([None])
        self.linenumber = 1
        if fileobj is None:
            fileobj = _open(filename)
        try:
            source = fileobj.read().decode('utf-8')
        except UnicodeDecodeError as e:
            print('Aborting due to parse error in %s: %s' %
                    (self.filename, e), file=sys.stderr)
            sys.exit(1)
        ElementProgram.__init__(self, source, filename=filename)
        return [m.message() if isinstance(m, TranslateContext) else m
                for m in self.messages]

    def visit(self, kind, args):
        visitor = getattr(self, 'visit_%s' % kind, None)
        if visitor is not None:
            return visitor(*args)
        else:
            print("Warning: Unknown node type '%s', linenumbers might be off. Please report this warning." %
                    kind, file=sys.stderr)

    def visit_start_tag(self, element):
        self.visit_element(element, None, [])

    def visit_element(self, start, end, children):
        self.linenumber += get_newline_count(start['prefix'] + start['name'])
        if self.translatestack and self.translatestack[-1]:
            self.translatestack[-1].add_element(start)

        attributes = start['ns_attrs']
        plain_attrs = get_plain_attrs(start['attrs'])
        childs_lineno = self.linenumber
        post_offset = [x[2] for x in plain_attrs.values()]
        if post_offset:
            childs_lineno += max(post_offset)
        childs_lineno += get_newline_count(start['suffix'])
        new_domain = attributes.get((I18N_NS, 'domain'))
        old_domain = self.domainstack[-1][0] if self.domainstack else None
        new_context = attributes.get((I18N_NS, 'context'))
        old_context = self.domainstack[-1][1] if self.domainstack else None
        comment = attributes.get((I18N_NS, 'comment'))
        if new_domain or new_context:
            self.domainstack.append((new_domain or old_domain, new_context or old_context))
        elif self.domainstack:
            self.domainstack.append(self.domainstack[-1])

        i18n_translate = attributes.get((I18N_NS, 'translate'))
        if i18n_translate is not None:
            ctx = TranslateContext(
                self.domainstack[-1][0] if self.domainstack else None,
                self.domainstack[-1][1] if self.domainstack else None,
                i18n_translate, comment, self.filename, childs_lineno)
            if self.translatestack:
                ctx.parent = self.translatestack[-1]
                if ctx.parent is not None:
                    ctx.parent.register_child(start, ctx)
            self.translatestack.append(ctx)
        else:
            self.translatestack.append(None)

        if self.domainstack:
            i18n_attributes = attributes.get((I18N_NS, 'attributes'))
            if i18n_attributes:
                parts = [p.strip() for p in i18n_attributes.split(';')]
                for msgid in parts:
                    if ' ' not in msgid:
                        if msgid not in plain_attrs:
                            continue
                        value, offset, post_offset = plain_attrs[msgid]
                        self.add_message(value, offset=offset)
                    else:
                        try:
                            (attr, msgid) = msgid.split()
                        except ValueError:
                            continue
                        if attr not in plain_attrs:
                            continue
                        value, offset, post_offset = plain_attrs[attr]
                        self.add_message(msgid, u'Default: %s' % value, offset=offset)

            for (attribute, value) in attributes.items():
                value = decode_htmlentities(value)
                for source in self.get_code_for_attribute(attribute, value):
                    self.parse_python(source)

        self.linenumber = childs_lineno
        for child in children:
            self.visit(*child)

        if end is not None:
            self.linenumber += get_newline_count(end['prefix'] + end['name'])
            post_offset = [x[2] for x in get_plain_attrs(end['attrs']).values()]
            if post_offset:
                self.linenumber += max(post_offset)
            self.linenumber += get_newline_count(end['suffix'])

        if self.domainstack:
            self.domainstack.pop()

        translate = self.translatestack.pop()
        if translate and not translate.ignore() and translate.domain and \
                (self.target_domain is None or translate.domain == self.target_domain):
            self.messages.append(translate)

    def visit_text(self, data):
        default_engine = self.config['default-engine']
        for line in data.splitlines():
            for source in get_python_expressions(line, default_engine):
                if UNDERSCORE_CALL.search(source):
                    self.parse_python(source)
        if self.translatestack[-1]:
            self.translatestack[-1].add_text(data)
        self.linenumber += get_newline_count(data)

    def visit_comment(self, data):
        self.linenumber += get_newline_count(data)

    def visit_cdata(self, data):
        self.linenumber += get_newline_count(data)

    def visit_default(self, data):
        if not data.lower().startswith('<!doctype'):
            print("%s:%s\n    Warning: Node type 'default', possible bad markup" %
                    (self.filename, self.linenumber), file=sys.stderr)
        self.linenumber += get_newline_count(data)

    def add_message(self, msgid, comment=u'', offset=0):
        self.messages.append(Message(None, msgid, None, [], comment, u'',
            (self.filename, self.linenumber + offset)))

    def _assert_valid_python(self, value):
        if not is_valid_python(value):
            print('Aborting due to Python syntax error in %s[%d]: %s' %
                    (self.filename, self.linenumber, value))
            sys.exit(1)

    def get_code_for_attribute(self, attribute, value):
        default_engine = self.config['default-engine']
        if attribute[0] == TAL_NS:
            if attribute[1] in ['content', 'replace']:
                for (engine, value) in split_expression(value, default_engine):
                    if engine == 'python':
                        m = STRUCTURE_PREFIX.match(value)
                        if m is not None:
                            value = m.group(2)
                        m = ENGINE_PREFIX.match(value)
                        if (m is not None) and (m.group(1) == 'python'):
                            m = None
                            value = value.split(':', 1)[1]
                        if m is None:
                            value = '(%s)' % value
                            self._assert_valid_python(value)
                            yield value
            if attribute[1] == 'define':
                for (scope, var, value) in parse_defines(value):
                    for (engine, value) in split_expression(value, default_engine):
                        if engine == 'python':
                            value = '(%s)' % value
                            self._assert_valid_python(value)
                            yield value
            elif attribute[1] == 'repeat':
                defines = parse_defines(value)
                if len(defines) != 1:
                    print('Aborting due to syntax error in %s[%d]: %s' % (
                            self.filename, self.linenumber, value))
                scope, var, value = defines[0]
                for (engine, value) in split_expression(value, default_engine):
                    if engine == 'python':
                        self._assert_valid_python(value)
                        yield value
        else:
            try:
                for source in get_python_expressions(value, default_engine):
                    yield source
            except SyntaxError:
                print('Aborting due to Python syntax error in %s[%d]: %s' %
                        (self.filename, self.linenumber, value))
                sys.exit(1)

    def parse_python(self, source):
        assert isinstance(source, type(u''))
        for message in _extract_python(self.filename, source, self.options, self.linenumber):
            self.messages.append(Message(*message[:6],
                location=(self.filename, self.linenumber + message.location[1])))


class ZopeExtractor(ChameleonExtractor):
    '''Zope templates (defaults to TALES expressions)'''
    extensions = ['.zpt', '.cpt']
    default_config = {
            'default-engine': 'tales',
            }


def is_valid_python(source):
    try:
        ast.parse(source, mode='eval')
    except SyntaxError:
        return False
    else:
        return True


def split_expression(source, default_engine):
    for part in split_parts.split(source):
        expression = part.strip().replace('\\|', '|')
        yield get_tales_engine(expression, default_engine)


def get_tales_engine(source, default_engine):
    m = ENGINE_PREFIX.match(source)
    if m is None:
        return (default_engine, source)
    else:
        return (m.group(1), source[m.end():])


def get_python_expressions(source, default_engine):
    regex = re.compile(r'(?<!\\)\$({(?P<expression>.*)})', re.DOTALL)
    while source:
        m = regex.search(source)
        if m is None:
            break

        source = source[m.start():]
        matched = m.group(0)

        # We foundsomething that looks like ${...}, but could also be
        # ${...}..}, so keep trying to parse while stripping the last
        # character until either all python validates, or we no longer
        # match

        m = regex.search(source)
        while m is not None:
            candidate = m.group('expression')
            candidates = [code for (engine, code) in split_expression(candidate, default_engine)
                          if engine == 'python']
            if all(is_valid_python(c) for c in candidates):
                # All valid, so return and move to next ${ block
                for c in candidates:
                    yield c
                source = source[m.end():]
                break
            else:
                # Syntax error somewhere, so try again with last character
                # stripped.
                matched = matched[:-1]
                m = regex.search(matched)

        if m is None:
            # We found ${, but could not find a valid python expression
            raise SyntaxError()



from pkg_resources import working_set
from . import EXTRACTORS
from . import Message
from . import check_c_format
from . import check_python_format
from . import Extractor


DEFAULT_KEYWORDS = {
        '_': None,
        'gettext': None,
        'ngettext': (1, 2),
        'ugettext': None,
        'ungettext': (1, 2),
        'dgettext': (2,),
        'dngettext': (2, 3),
        'N_': None,
        'pgettext': ((1, 'c'), 2)
        }


class BabelExtractor(Extractor):
    extensions = []
    extractor = None
    default_config = {
            'comment-tags': '',
    }

    def __call__(self, filename, options, fileobj=None, lineno=0):
        if fileobj is None:
            fileobj = open(filename, 'rb')
        comment_tags = self.config['comment-tags'].split()

        for (lineno, _, msgid, comment) in self.extractor(fileobj, DEFAULT_KEYWORDS.keys(),
                comment_tags, self.config):
            if isinstance(msgid, tuple):
                (msgid, msgid_plural) = msgid[:2]
            else:
                msgid_plural = u''
            comment = u' '.join(comment)
            flags = []
            check_c_format(msgid, flags)
            check_python_format(msgid, flags)
            yield Message(None, msgid, msgid_plural, flags, comment, u'', (filename, lineno))


def register_babel_plugins():
    for entry_point in working_set.iter_entry_points('babel.extractors'):
        name = entry_point.name
        extractor = entry_point.load(require=True)
        cls = type('BabelExtractor_%s' % name,
                (BabelExtractor, object),
                {'extractor': staticmethod(extractor),
                 '__doc__': extractor.__doc__.splitlines()[0]})
        EXTRACTORS['babel-%s' % name] = cls()



# coding=utf-8
import mock
import pytest
from lingua.extractors.python import PythonExtractor


python_extractor = PythonExtractor()
source = None


@pytest.fixture
def fake_source(request):
    patcher = mock.patch('lingua.extractors.python._read',
            side_effect=lambda *a: source)
    patcher.start()
    request.addfinalizer(patcher.stop)


@pytest.mark.usefixtures('fake_source')
def test_syntax_error():
    global source
    options = mock.Mock()
    options.keywords = []
    source = u'''def class xya _(u'føo')'''
    with pytest.raises(SystemExit):
        generator = python_extractor('filename', options)
        list(generator)


@pytest.mark.usefixtures('fake_source')
def test_multiline_string():
    global source
    options = mock.Mock()
    options.keywords = []
    source = u'''_(u'őne two '\n'three')'''
    messages = list(python_extractor('filename', options))
    assert len(messages) == 1
    assert messages[0].msgid == u'őne two three'


@pytest.mark.usefixtures('fake_source')
def test_plural():
    global source
    options = mock.Mock()
    options.keywords = []
    source = u'''ngettext(u'one côw', u'%d cows', 5)'''
    messages = list(python_extractor('filename', options))
    assert len(messages) == 1
    assert messages[0].msgid == u'one côw'
    assert messages[0].msgid_plural == u'%d cows'


@pytest.mark.usefixtures('fake_source')
def test_translationstring_parameters():
    global source
    options = mock.Mock()
    options.keywords = []
    source = u'''_(u'msgid', default=u'Canonical text')'''
    messages = list(python_extractor('filename', options))
    assert len(messages) == 1
    assert messages[0].msgctxt is None
    assert messages[0].msgid == u'msgid'
    assert messages[0].comment == u'Default: Canonical text'


@pytest.mark.usefixtures('fake_source')
def test_translationstring_context():
    global source
    options = mock.Mock()
    options.keywords = []
    source = u'''_(u'Canonical text', context='button')'''
    messages = list(python_extractor('filename', options))
    assert len(messages) == 1
    assert messages[0].msgctxt == 'button'
    assert messages[0].msgid == u'Canonical text'


@pytest.mark.usefixtures('fake_source')
def test_function_call_in_keyword():
    global source
    options = mock.Mock()
    options.keywords = ['other']
    source = u'''other(six.u('word'))'''
    messages = list(python_extractor('filename', options))
    assert len(messages) == 0


@pytest.mark.usefixtures('fake_source')
def test_use_lineno_parameter():
    global source
    options = mock.Mock()
    options.keywords = []
    source = u'''_(u'word')'''
    messages = list(python_extractor('filename', options, lineno=5))
    assert len(messages) == 1
    assert messages[0].location[1] == 6



from __future__ import print_function
import argparse
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict
from operator import attrgetter
import os
import sys
import time
try:
    from configparser import SafeConfigParser
except ImportError:
    from ConfigParser import SafeConfigParser
import polib
from lingua.extractors import get_extractor
from lingua.extractors import register_extractors
from lingua.extractors.babel import register_babel_plugins
from lingua.extractors import EXTRACTORS
from lingua.extractors import EXTENSIONS
from lingua import lingua_version


def po_timestamp():
    local = time.localtime()
    offset = -(time.altzone if local.tm_isdst else time.timezone)
    return '%s%s%s' % (
        time.strftime('%Y-%m-%d %H:%M', local),
        '-' if offset < 0 else '+',
        time.strftime('%H%M', time.gmtime(abs(offset))))


class POEntry(polib.POEntry):
    def __init__(self, *a, **kw):
        polib.POEntry.__init__(self, *a, **kw)
        self._comments = []
        self._tcomments = []

    @property
    def comment(self):
        return u'\n'.join(self._comments)

    @comment.setter
    def comment(self, value):
        pass

    @property
    def tcomment(self):
        return u'\n'.join(self._tcomments)

    @tcomment.setter
    def tcomment(self, value):
        pass

    def update(self, message):
        self.occurrences.append(message.location)
        self.flags.extend(f for f in message.flags if f not in self.flags)
        if message.comment not in self._comments:
            self._comments.append(message.comment)
        if message.tcomment not in self._tcomments:
            self._tcomments.append(message.tcomment)


class POFile(polib.POFile):
    copyright = None
    package_name = None

    def metadata_as_entry(self):
        entry = polib.POFile.metadata_as_entry(self)
        year = time.localtime().tm_year
        header = [u'SOME DESCRIPTIVE TITLE']
        if self.copyright_holder:
            header.append(u'Copyright (C) %d %s' % (year, self.copyright_holder))
        header.append(
                u'This file is distributed under the same license as the %s package.' %
                self.package_name)
        header.append(u'FIRST AUTHOR <EMAIL@ADDRESS>, %d.' % year)
        entry.tcomment = u'\n'.join(header)
        return entry


def no_duplicates(iterator):
    seen = set()
    for item in iterator:
        if item in seen:
            continue
        seen.add(item)
        yield item


def list_files(options):
    if options.files_from:
        for filename in open(options.files_from, 'r'):
            if filename.startswith('#') or not filename.strip():
                continue
            yield filename
    for file in options.file:
        if os.path.isfile(file):
            yield file
        elif os.path.isdir(file):
            for (dirpath, dirnames, filenames) in os.walk(file):
                for file in filenames:
                    if get_extractor(file) is not None:
                        yield os.path.join(dirpath, file)
        else:
            print('Invalid file type for %s' % file, file=sys.stderr)
            sys.exit(1)


def find_file(filename, search_path=[]):
    """Return the filename for a given file, checking search paths.
    """
    paths = [os.path.curdir] + search_path
    for path in paths:
        filename = os.path.join(path, filename)
        if os.path.isfile(filename):
            return filename
    return None


def create_catalog(options):
    catalog = POFile(wrapwidth=options.width)
    catalog.copyright_holder = options.copyright_holder
    catalog.package_name = options.package_name
    catalog.metadata_is_fuzzy = True
    catalog.metadata = OrderedDict()
    catalog.metadata['Project-Id-Version'] = ' '.join(filter(
        None, [options.package_name, options.package_version]))
    if options.msgid_bugs_address:
        catalog.metadata['Report-Msgid-Bugs-To'] = options.msgid_bugs_address
    catalog.metadata['POT-Creation-Date'] = po_timestamp()
    catalog.metadata['PO-Revision-Date'] = 'YEAR-MO-DA HO:MI+ZONE'
    catalog.metadata['Last-Translator'] = 'FULL NAME <EMAIL@ADDRESS'
    catalog.metadata['Language-Team'] = 'LANGUAGE <LL@li.org>'
    catalog.metadata['Language'] = ''
    catalog.metadata['MIME-Version'] = '1.0'
    catalog.metadata['Content-Type'] = 'text/plain; charset=UTF-8'
    catalog.metadata['Content-Transfer-Encoding'] = '8bit'
    catalog.metadata['Generated-By'] = 'Lingua %s' % lingua_version()
    return catalog


def _register_extension(extension, extractor):
    if extractor not in EXTRACTORS:
        print('Unknown extractor %s. Check --list-extractors for available options' % extractor,
                file=sys.stderr)
        sys.exit(1)
    EXTENSIONS[extension] = extractor


def read_config(filename):
    config = SafeConfigParser()
    config.readfp(open(filename))
    for section in config.sections():
        if section == 'extensions':
            for (extension, extractor) in config.items(section):
                _register_extension(extension, extractor)
        elif section.startswith('extractor:'):
            extractor = section[10:]
            if extractor not in EXTRACTORS:
                print('Unknown extractor %s. '
                      'Check --list-extractors for available options' % extractor,
                        file=sys.stderr)
                sys.exit(1)
            config = dict(config.items(section))
            EXTRACTORS[extractor].update_config(**config)
        elif section.startswith('extension'):
            print('Use of %s section is obsolete. '
                  'Please use the "extensions" section.' % section,
                  file=sys.stderr)
            extension = section[10:]
            plugin = config.get(section, 'plugin')
            if not plugin:
                print('No plugin defined for extension %s' % extension,
                    file=sys.stderr)
            _register_extension(extension, plugin)


def main():
    parser = argparse.ArgumentParser(
            description='Extract translateable strings.')

    parser.add_argument('-c', '--config', metavar='CONFIG',
            help='Read configuration from CONFIG file')
    # Input options
    parser.add_argument('-f', '--files-from', metavar='FILE',
            help='Get list of files to process from FILE')
    parser.add_argument('-D', '--directory', metavar='DIRECTORY',
            action='append', default=[],
            help='Add DIRECTORY to list of paths to check for input files')
    parser.add_argument('file', nargs='*',
            help='Source file to process')
    parser.add_argument('--list-extractors', action='store_true',
            help='List all known extraction plugins')
    # Output options
    parser.add_argument('-o', '--output', metavar='FILE',
            default='messages.pot',
            help='Filename for generated POT file')
    parser.add_argument('--no-location',
            action='store_false', dest='location',
            help='Do not include location information')
    parser.add_argument('-n', '--add-location',
            action='store_true', dest='location', default=True,
            help='Include location information (default)')
    parser.add_argument('-w', '--width', metavar='NUMBER',
            default=79,
            help='Output width')
    parser.add_argument('-s', '--sort-output', # babel compatibility
            action='store_const', const='msgid', dest='sort',
            help='Order messages by their msgid')
    parser.add_argument('-F', '--sort-by-file',
            action='store_const', const='location', dest='sort',
            help='Order messages by file location')
    # Extraction configuration
    parser.add_argument('-d', '--domain',
            help='Domain to extract')
    parser.add_argument('-k', '--keyword', metavar='WORD',
            dest='keywords', action='append', default=[], nargs='?',
            help='Look for WORD as additional keyword')
    # POT metadata
    parser.add_argument('--copyright-holder', metavar='STRING',
            help='Specifies the copyright holder for the texts')
    parser.add_argument('--package-name', metavar='NAME',
            default=u'PACKAGE',
            help='Package name to use in the generated POT file')
    parser.add_argument('--package-version', metavar='Version',
            default='1.0',
            help='Package version to use in the generated POT file')
    parser.add_argument('--msgid-bugs-address', metavar='EMAIL',
            help='Email address bugs should be send to')

    options = parser.parse_args()
    register_extractors()
    register_babel_plugins()

    if options.list_extractors:
        for extractor in sorted(EXTRACTORS):
            print('%-17s %s' % (extractor, EXTRACTORS[extractor].__doc__ or ''))
        return

    if options.config:
        read_config(options.config)
    else:
        user_home = os.path.expanduser('~')
        global_config = os.path.join(user_home, '.config', 'lingua')
        if os.path.exists(global_config):
            read_config(global_config)

    catalog = create_catalog(options)

    scanned = 0
    for filename in no_duplicates(list_files(options)):
        real_filename = find_file(filename, options.directory)
        if real_filename is None:
            print('Can not find file %s' % filename, file=sys.stderr)
            sys.exit(1)
        extractor = get_extractor(real_filename)
        if extractor is None:
            print('No extractor available for file %s' % filename, file=sys.stderr)
            sys.exit(1)

        for message in extractor(real_filename, options):
            entry = catalog.find(message.msgid, msgctxt=message.msgctxt)
            if entry is None:
                entry = POEntry(msgctxt=message.msgctxt,
                                msgid=message.msgid)
                if message.msgid_plural:
                    entry.msgid_plural = message.msgid_plural
                    entry.msgstr_plural[0] = ''
                    entry.msgstr_plural[1] = ''
                catalog.append(entry)
            entry.update(message)
        scanned += 1
    if not scanned:
        print('No files scanned, aborting', file=sys.stderr)
        sys.exit(1)
    if not catalog:
        print('No translatable strings found, aborting', file=sys.stderr)
        sys.exit(2)

    if options.sort == 'msgid':
        catalog.sort(key=attrgetter('msgid'))
    elif options.sort == 'location':
        # Order the occurrences themselves, so the output is consistent
        catalog.sort(key=lambda m: m.occurrences.sort() or m.occurrences)

    catalog.save(options.output)


if __name__ == '__main__':
    main()



from lingua.extractors import check_c_format
from lingua.extractors import Keyword
from lingua.extractors import Extractor
import pytest


def test_no_format():
    flags = []
    check_c_format('Hello, world', flags)
    assert flags == []


def test_space_in_format():
    # This is technically a violation: printf(3) allows a space in format
    # strings, but in the real world this leads to many false positives.
    flags = []
    check_c_format('This is 5% of everything', flags)
    assert flags == []


def test_basic_c_format():
    flags = []
    check_c_format('Hello, %s', flags)
    assert 'c-format' in flags


def test_accept_no_format_hint():
    flags = ['no-c-format']
    check_c_format('Hello, %s', flags)
    assert 'c-format' not in flags


def test_escaped_percent_sign():
    flags = []
    check_c_format('100%%', flags)
    assert 'c-format' not in flags


def test_strftime_is_not_c_format():
    flags = []
    check_c_format('%Y-%m-%d', flags)
    assert 'c-format' not in flags


class TestKeywordFromSpec(object):
    def test_minimal(self):
        kw = Keyword.from_spec('gettext')
        assert kw.function == 'gettext'
        assert kw.msgid_param == 1
        assert kw.msgid_plural_param is None

    def test_custom_msgid_param(self):
        kw = Keyword.from_spec('i18n_log:2')
        assert kw.function == 'i18n_log'
        assert kw.msgid_param == 2
        assert kw.msgid_plural_param is None

    def test_plural(self):
        kw = Keyword.from_spec('ngettext:1,2')
        assert kw.function == 'ngettext'
        assert kw.msgid_param == 1
        assert kw.msgid_plural_param == 2

    def test_domain_param(self):
        kw = Keyword.from_spec('dgettext:1d,2')
        assert kw.function == 'dgettext'
        assert kw.msgid_param == 2
        assert kw.msgid_plural_param is None
        assert kw.domain_param == 1

    def test_parameter_count(self):
        kw = Keyword.from_spec('myfunc:1,5t')
        assert kw.function == 'myfunc'
        assert kw.msgid_param == 1
        assert kw.msgid_plural_param is None
        assert kw.required_arguments == 5


def test_extractor():
    with pytest.raises(TypeError):
        Extractor()



from __future__ import print_function
from pkg_resources import working_set
import abc
import collections
import os
import re
import sys
from .compat import add_metaclass


Message = collections.namedtuple('Message',
        'msgctxt msgid msgid_plural flags comment tcomment location')

EXTRACTORS = {}
EXTENSIONS = {}


def get_extractor(filename):
    ext = os.path.splitext(filename)[1]
    try:
        return EXTRACTORS[EXTENSIONS[ext]]
    except KeyError:
        return None


# Based on http://www.cplusplus.com/reference/cstdio/printf/
# Note that we skip the space-flag in this list, since this creates too
# many false positives.
_C_FORMAT = re.compile(r'''
        %
        [+#0-]?               # flags
        (\d+|\*)?             # width
        (\.(\d+|\*))?         # precision
        (hh?|ll?|j|z|t|L)?    # length
        [diuoxXfFeEgGaAcspn%] # specifier
        ''', re.VERBOSE)


def check_c_format(buf, flags):
    if 'no-c-format' in flags:
        return
    formats = list(re.finditer('%(?!%)', buf))
    if formats and all(_C_FORMAT.match(buf[m.start():]) is not None
            for m in formats):
        flags.append('c-format')


# Based on http://docs.python.org/2/library/string.html#format-string-syntax
_PYTHON_FORMAT = re.compile(r'''
        \{
            (([_A-Za-z](\w*)(\.[_a-z]\w*|\[\d+\])?)|\w+)?  # fieldname
            (![rs])?  # conversion
            (:\.?[<>=^]?[+ -]?\w*,?(\.\w+)?[bcdeEfFgGnosxX%]?)?  # format_spec
        \}
        ''', re.VERBOSE)


def check_python_format(buf, flags):
    if 'no-python-format' in flags:
        return
    if _PYTHON_FORMAT.search(buf) is not None:
        flags.append('python-format')


class Keyword(object):
    msgctxt_param = None
    domain_param = None
    comment = u''
    required_arguments = None

    _comment_arg = re.compile(r'^"(.*)"$')

    def __init__(self, function, msgid_param=1, msgid_plural_param=None, domain_param=None):
        self.function = function
        self.msgid_param = msgid_param
        self.msgid_plural_param = msgid_plural_param
        self.domain_param = domain_param

    @classmethod
    def from_spec(cls, spec):
        if ':' not in spec:
            return cls(spec)
        try:
            (function, args) = spec.split(':', 1)
            kw = cls(function)
            seen_msgid_param = False
            while args:
                if cls._comment_arg.match(args) is not None:
                    kw.comment = args[1:-1]
                    break
                (param, args) = args.split(',', 1) if ',' in args else (args, '')
                if param.endswith('c'):
                    kw.msgctxt_param = int(param[:-1])
                elif param.endswith('d'):
                    kw.domain_param = int(param[:-1])
                elif param.endswith('t'):
                    kw.required_arguments = int(param[:-1])
                elif not seen_msgid_param:
                    kw.msgid_param = int(param)
                    seen_msgid_param = True
                else:
                    kw.msgid_plural_param = int(param)
        except SyntaxError:
            raise ValueError('Invalid keyword spec: %s' % spec)
        return kw


def update_keywords(keywords, specs):
    for spec in specs:
        if not spec:
            keywords.clear()
        try:
            kw = Keyword.from_spec(spec)
        except ValueError as e:
            print(e, file=sys.stderr)
            sys.exit(1)
        keywords[kw.function] = kw


@add_metaclass(abc.ABCMeta)
class Extractor(object):
    default_config = {}

    def __init__(self, config=None):
        self.config = self.default_config.copy()
        if config:
            self.config.update(config)

    def update_config(self, **kw):
        self.config.update(kw)

    @abc.abstractproperty
    def extensions(self):
        raise NotImplemented()

    @abc.abstractmethod
    def __call__(self, filename, options, fileobj=None, lineno=0):
        raise NotImplemented()


def register_extractors():
    for entry_point in working_set.iter_entry_points('lingua.extractors'):
        extractor = entry_point.load(require=True)
        if not issubclass(extractor, Extractor):
            raise ValueError(
                u'Registered extractor must derive from ``Extractor``')
        EXTRACTORS[entry_point.name] = extractor()
        for extension in extractor.extensions:
            EXTENSIONS[extension] = entry_point.name



from __future__ import print_function
import ast
import sys
from . import Extractor
from . import Message
from . import check_c_format
from . import check_python_format
from . import Keyword
from . import update_keywords


KEYWORDS = {
        'gettext': Keyword('gettext'),
        'ugettext': Keyword('ugettext'),
        'dgettext': Keyword('dgettext', 2, domain_param=1),
        'ldgettext': Keyword('ldgettext', 2, domain_param=1),
        'ngettext': Keyword('ngettext', 1, 2),
        'lngettext': Keyword('ngettext', 1, 2),
        'ungettext': Keyword('ungettext', 1, 2),
        'dngettext': Keyword('dngettext', 2, 3, domain_param=1),
        'ldngettext': Keyword('dngettext', 2, 3, domain_param=1),
        }


def parse_keyword(node, keyword, filename, firstline):
    if keyword.required_arguments and len(node.args) != keyword.required_arguments:
        return None

    def get_string(param, error_msg):
        if not param:
            return None
        arg = node.args[param - 1]
        if not isinstance(node.args[0], ast.Str):
            print('%s[%d]: %s' %  (filename, firstline + arg.lineno, error_msg),
                    file=sys.stderr)
            raise IndexError()
        else:
            return arg.s

    try:
        domain = get_string(keyword.domain_param, 'Domain argument must be a string')
        msgid = get_string(keyword.msgid_param, 'Message argument must be a string')
        msgid_plural = get_string(keyword.msgid_plural_param, 'Plural message argument must be a string')
        msgctxt = get_string(keyword.msgctxt_param, 'Context argument must be a string')
        comment = keyword.comment
    except IndexError:
        return None
    return (domain, msgctxt, msgid, msgid_plural, comment)


def parse_translationstring(node):
    if not node.args:
        return None

    msgid = None
    context = None
    default = u''
    if isinstance(node.args[0], ast.Str):
        msgid = node.args[0].s
    if len(node.args) > 2 and isinstance(node.args[2], ast.Str):
        default = node.args[2].s
    for keyword in node.keywords:
        if not isinstance(keyword.value, ast.Str):
            continue
        if keyword.arg == 'msgid':
            msgid = keyword.value.s
        elif keyword.arg == 'default':
            default = keyword.value.s
        elif keyword.arg == 'context':
            context = keyword.value.s
    if not msgid:
        return None

    comment = u'Default: %s' % default if default else u''
    return (None, context, msgid, None, comment)


def _read(filename):
    """Injection point for tests."""
    return open(filename, 'rb').read()


def _extract_python(filename, source, options, firstline=0):
    update_keywords(KEYWORDS, options.keywords)
    try:
        tree = ast.parse(source, filename)
    except SyntaxError as e:
        print('Aborting due to parse error in %s[%d]: %s' %
                        (filename, firstline + e.lineno, e.text), file=sys.stderr)
        sys.exit(1)

    for node in ast.walk(tree):
        if not isinstance(node, ast.Call):
            continue
        if not isinstance(node.func, ast.Name):
            continue
        msg = None
        if node.func.id in KEYWORDS:
            msg = parse_keyword(node, KEYWORDS[node.func.id], filename, firstline)
        elif node.func.id == '_':
            msg = parse_translationstring(node)
        if msg is None:
            continue

        if options.domain is not None and msg[0] and msg[0] != options.domain:
            continue

        flags = []
        check_c_format(msg[2], flags)
        check_python_format(msg[2], flags)
        yield Message(msg[1], msg[2], msg[3], flags, msg[4], u'', (filename, firstline + node.lineno))


class PythonExtractor(Extractor):
    '''Python sources'''
    extensions = ['.py']

    def __call__(self, filename, options, fileobj=None, lineno=0):
        if fileobj is None:
            code = _read(filename)
        else:
            code = fileobj.read()
        return _extract_python(filename, code, options, lineno)



# Python compatibility support code
# This is taken from six


def add_metaclass(metaclass):
    """Class decorator for creating a class with a metaclass."""
    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper



import mock
import pytest
from io import BytesIO
from lingua.extractors.zcml import ZCMLExtractor


zcml_extractor = ZCMLExtractor()
source = None


def _options(**kw):
    options = mock.Mock()
    options.domain = None
    options.keywords = []
    for (k, w) in kw.items():
        setattr(options, k, w)
    return options


@pytest.fixture
def fake_source(request):
    patcher = mock.patch('lingua.extractors.zcml._open',
            side_effect=lambda *a: BytesIO(source))
    patcher.start()
    request.addfinalizer(patcher.stop)


@pytest.mark.usefixtures('fake_source')
def test_empty_zcml():
    global source
    source = b'''<configure/>'''
    assert list(zcml_extractor('filename', _options())) == []


@pytest.mark.usefixtures('fake_source')
def test_i18n_without_domain():
    global source
    source = b'''\
                <configure>
                  <dummy title="test title"/>
                </configure>
              '''
    assert list(zcml_extractor('filename', _options())) == []


@pytest.mark.usefixtures('fake_source')
def test_i18n_with_domain():
    global source
    source = b'''\
                <configure i18n_domain="lingua">
                  <dummy title="test title"/>
                </configure>
              '''
    messages = list(zcml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'test title'


@pytest.mark.usefixtures('fake_source')
def test_multiple_messages():
    global source
    source = b'''\
                <configure i18n_domain="lingua">
                  <dummy title="test title 1"/>
                  <dummy title="test title 2"/>
                </configure>
              '''
    messages = list(zcml_extractor('filename', _options()))
    assert len(messages) == 2
    assert messages[0].msgid == u'test title 1'
    assert messages[1].msgid == u'test title 2'


@pytest.mark.usefixtures('fake_source')
def test_domain_nesting():
    global source
    source = b'''\
                <configure>
                  <configure i18n_domain="lingua">
                      <dummy title="test title 1"/>
                  </configure>
                  <dummy title="test title 2"/>
                </configure>
              '''
    messages = list(zcml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'test title 1'


@pytest.mark.usefixtures('fake_source')
def test_abort_on_syntax_error():
    global source
    source = b'''<configure'''
    with pytest.raises(SystemExit):
        list(zcml_extractor('filename', _options()))



from setuptools import setup, find_packages
from setuptools.command.test import test as TestCommand
import sys

version = '3.11'

install_requires = [
        'setuptools',
        'polib',
        'Chameleon',
        ]
if sys.version_info < (2, 7):
    install_requires.append('argparse')
    install_requires.append('ordereddict')

tests_require = [
        'pytest',
        'mock',
        ]


class PyTest(TestCommand):
    def finalize_options(self):
        TestCommand.finalize_options(self)
        self.test_args = ['tests']
        self.test_suite = True

    def run_tests(self):
        # import here, cause outside the eggs aren't loaded
        import pytest
        errno = pytest.main(self.test_args)
        sys.exit(errno)


setup(name='lingua',
      version=version,
      description='Translation toolset',
      long_description=open('README.rst').read() + '\n' +
              open('changes.rst').read(),
      classifiers=[
          'Intended Audience :: Developers',
          'License :: DFSG approved',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.3',
          'Programming Language :: Python :: 3.4',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],
      keywords='translation po gettext Babel',
      author='Wichert Akkerman',
      author_email='wichert@wiggy.net',
      url='https://github.com/wichert/lingua',
      license='BSD',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      include_package_data=True,
      zip_safe=True,
      install_requires=install_requires,
      tests_require=tests_require,
      extras_require={'tests': tests_require},
      cmdclass={'test': PyTest},
      entry_points='''
      [console_scripts]
      polint = lingua.polint:main
      pot-create = lingua.extract:main

      [lingua.extractors]
      python = lingua.extractors.python:PythonExtractor
      chameleon = lingua.extractors.xml:ChameleonExtractor
      xml = lingua.extractors.xml:ChameleonExtractor
      zope = lingua.extractors.xml:ZopeExtractor
      zcml = lingua.extractors.zcml:ZCMLExtractor
      '''
      )



# coding=utf-8
import mock
import pytest
from io import BytesIO
from lingua.extractors.xml import ChameleonExtractor
from lingua.extractors.xml import get_python_expressions


xml_extractor = ChameleonExtractor()
source = None


def _options(**kw):
    options = mock.Mock()
    options.domain = None
    options.keywords = []
    for (k, w) in kw.items():
        setattr(options, k, w)
    return options


@pytest.fixture
def fake_source(request):
    patcher = mock.patch('lingua.extractors.xml._open',
            side_effect=lambda *a: BytesIO(source))
    patcher.start()
    request.addfinalizer(patcher.stop)


@pytest.mark.usefixtures('fake_source')
def test_abort_on_syntax_error():
    global source
    source = b'''\xff\xff\xff'''
    with pytest.raises(SystemExit):
        list(xml_extractor('filename', _options()))


@pytest.mark.usefixtures('fake_source')
def test_empty_xml():
    global source
    source = b'''<html/>'''
    assert list(xml_extractor('filename', _options())) == []


@pytest.mark.usefixtures('fake_source')
def test_attributes_plain():
    global source
    source = u'''<html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:attributes="title" title="tést title"/>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'tést title'
    assert messages[0].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_custom_i18n_namespace():
    global source
    source = b'''<html i18n:domain="other">
                   <dummy i18n:translate="">Foo</dummy>
                   <other xmlns:i="http://xml.zope.org/namespaces/i18n"
                          i:domain="lingua">
                     <dummy i:translate="">Foo</dummy>
                   </other>
                   <dummy i18n:translate="">Foo</dummy>
                   <dummy i18n:translate="">Foo</dummy>
                 </html>
                 '''
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 4
    assert [m.msgid for m in messages] == [u'Foo'] * 4
    assert [m.location[1] for m in messages] == [2, 5, 7, 8]


@pytest.mark.usefixtures('fake_source')
def test_attributes_explicit_MessageId():
    global source
    source = u'''<html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                       i18n:domain="lingua">
                   <dummy i18n:attributes="title msg_title" title="test tïtle"/>
                 </html>
                  '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == 'msg_title'
    assert messages[0].comment == u'Default: test tïtle'
    assert messages[0].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_attributes_no_domain_without_domain_filter():
    global source
    source = b'''<html xmlns:i18n="http://xml.zope.org/namespaces/i18n">
                   <dummy i18n:attributes="title" title="test title"/>
                  </html>'''
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].location[1] == 2


@pytest.mark.usefixtures('fake_source')
def test_attributes_multiple_attributes():
    global source
    source = u'''<html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                       i18n:domain="lingua">
                   <dummy i18n:attributes="title ; alt" title="tést title"
                          alt="test ålt"/>
                 </html>'''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 2
    assert [m.msgid for m in messages] == [u'tést title', u'test ålt']
    assert [m.location[1] for m in messages] == [3, 4]


@pytest.mark.usefixtures('fake_source')
def test_attributes_multiple_attributes_explicit_msgid():
    global source
    source = u'''<html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:attributes="title msg_title; alt msg_alt"
                         title="test titlé" alt="test ålt"/>
                </html>'''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 2
    assert messages[0].msgid == u'msg_title'
    assert messages[0].comment == u'Default: test titlé'
    assert messages[1].msgid == u'msg_alt'
    assert messages[1].comment == u'Default: test ålt'
    assert [m.location[1] for m in messages] == [4, 4]


@pytest.mark.usefixtures('fake_source')
def test_translate_minimal():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:translate="">Dummy téxt</dummy>
                </html>'''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'Dummy téxt'
    assert messages[0].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_translate_explicit_msgid():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:translate="msgid_dummy">Dummy téxt</dummy>
                </html>'''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'msgid_dummy'
    assert messages[0].comment == u'Default: Dummy téxt'
    assert messages[0].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_translate_subelement():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:translate="msgid_dummy">Dummy
                    <strong>text</strong> demø</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'msgid_dummy'
    assert messages[0].comment == u'Default: Dummy <dynamic element> demø'
    assert messages[0].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_translate_named_subelement():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:translate="msgid_dummy">Dummy
                    <strong i18n:name="text">téxt</strong> demø</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'msgid_dummy'
    assert messages[0].comment == u'Default: Dummy ${text} demø'
    assert messages[0].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_translate_translated_subelement():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:translate="">Dummy
                    <strong i18n:name="text"
                            i18n:translate="">téxt</strong>
                    demø</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 2
    assert messages[0].msgid == u'téxt'
    assert messages[0].comment == u'Used in sentence: "Dummy ${text} demø"'
    assert messages[0].location[1] == 5
    assert messages[1].msgid == u'Dummy ${text} demø'
    assert messages[1].comment == u'Canonical text for ${text} is: "téxt"'
    assert messages[1].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_translate_translated_subelement_with_id():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:translate="msgid_dummy">Dummy
                    <strong i18n:name="text"
                            i18n:translate="msgid_text">téxt</strong>
                    demø</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 2
    assert messages[0].msgid == u'msgid_text'
    assert messages[0].comment == u'Default: téxt\nUsed in sentence: "Dummy ${text} demø"'
    assert messages[0].location[1] == 5
    assert messages[1].msgid == u'msgid_dummy'
    assert messages[1].comment == u'Default: Dummy ${text} demø\n' \
                                  u'Canonical text for ${text} is: "téxt"'
    assert messages[1].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_strip_extra_whitespace():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:translate="">Dummy


                  text</dummy>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'Dummy text'
    assert messages[0].location[1] == 3


@pytest.mark.usefixtures('fake_source')
def test_strip_trailing_and_leading_whitespace():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy i18n:translate="">
                    Dummy text
                  </dummy>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'Dummy text'


@pytest.mark.usefixtures('fake_source')
def test_html_entities():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <button i18n:translate="">Lock &amp; load&nbsp;</button>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'Lock &amp; load&nbsp;'


@pytest.mark.usefixtures('fake_source')
def test_ignore_undeclared_namespace():
    global source
    source = b'''\
                <html xmlns="http://www.w3.org/1999/xhtml"
                      xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <tal:block/>
                  <p i18n:translate="">Test</p>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'Test'


@pytest.mark.usefixtures('fake_source')
def test_ignore_dynamic_message():
    global source
    source = b'''\
                <html xmlns="http://www.w3.org/1999/xhtml"
                      xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <p i18n:translate="">${'dummy'}</p>
                </html>
                '''
    assert list(xml_extractor('filename', _options())) == []


@pytest.mark.usefixtures('fake_source')
def test_translate_call_in_python_expression_attribute():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy tal:replace="_(u'foo')">Dummy</dummy>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'foo'


@pytest.mark.usefixtures('fake_source')
def test_translate_call_in_python_expression_repeat_attribute():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy tal:repeat="label _(u'foo')">${label}</dummy>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'foo'


@pytest.mark.usefixtures('fake_source')
def test_translate_call_in_python_expression_in_char_data():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy>${_(u'foo')}</dummy>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'foo'


@pytest.mark.usefixtures('fake_source')
def test_translate_call_in_python_expression_in_attribute():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy title="${_(u'foo')}"></dummy>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'foo'


@pytest.mark.usefixtures('fake_source')
def test_multiple_expressions_with_translate_calls():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy title="${_(u'foo')} ${_(u'bar')}"></dummy>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'foo'
    assert messages[1].msgid == u'bar'


@pytest.mark.usefixtures('fake_source')
def test_translate_multiple_defines():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <tal:analytics define="isAnon _('one'); account _('two')">
                  </tal:analytics>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 2
    assert messages[0].msgid == u'one'
    assert messages[1].msgid == u'two'


@pytest.mark.usefixtures('fake_source')
def test_translate_explicit_python_expression_engine():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <tal:analytics define="layout python:_('one'); account _('two')">
                  </tal:analytics>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 2
    assert messages[0].msgid == u'one'
    assert messages[1].msgid == u'two'


@pytest.mark.usefixtures('fake_source')
def test_translate_ignore_other_expression_engine():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <tal:analytics define="layout load:layout.pt; account _('two')">
                  </tal:analytics>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'two'


@pytest.mark.usefixtures('fake_source')
def test_translate_ignore_other_content_expression_engine_with_structure():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <tal:var content="structure provider:my.content.provider">
                  </tal:var>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 0


@pytest.mark.usefixtures('fake_source')
def test_translate_explicit_python_engine_with_structure():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <tal:var content="structure python:_('foo')">
                  </tal:var>
                </html>
                '''
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'foo'


@pytest.mark.usefixtures('fake_source')
def test_translate_entities_in_python_expression():
    global source
    source = b'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <div class="${' disabled' if 1 &lt; 2 else None}"/>
                </html>
                '''
    list(xml_extractor('filename', _options()))


@pytest.mark.usefixtures('fake_source')
def test_curly_brace_in_python_expression():
    global source
    source = b'''\
            <html>
              <p>${request.route_url('set_locale', _query={'locale': 'de'})}</p>
            </html>
            '''
    list(xml_extractor('filename', _options()))


@pytest.mark.usefixtures('fake_source')
def test_curly_brace_in_python_attribute_expression():
    global source
    source = b'''\
            <html>
              <a href="${request.route_url('set_locale', _query={'locale': 'de'})}"></a>
            </html>
            '''
    list(xml_extractor('filename', _options()))


@pytest.mark.usefixtures('fake_source')
def test_curly_brace_related_syntax_error():
    global source
    source = b'''\
            <html>
              <a href="${request.route_url('set_locale', _}"></a>
            </html>
            '''
    with pytest.raises(SystemExit):
        list(xml_extractor('filename', _options()))


class Test_get_python_expression(object):
    def test_no_expressions(self):
        assert list(get_python_expressions('no python here', 'python')) == []

    def test_single_expression(self):
        assert list(get_python_expressions('${some_python}', 'python')) == ['some_python']

    def test_two_expressions(self):
        assert list(get_python_expressions('${one} ${two}', 'python')) == ['one', 'two']

    def test_nested_braces(self):
        assert list(get_python_expressions(
            '''${resource_url(_query={'one': 'one'})}''', 'python')) == \
            ['''resource_url(_query={'one': 'one'})''']


@pytest.mark.usefixtures('fake_source')
def test_python_expression_in_tales_expressions():
    global source
    source = u'''
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy tal:define="css_class css_class|string:${field.widget.css_class};">Dummy</dummy>
                </html>
                '''.encode('utf-8')
    assert list(xml_extractor('filename', _options())) == []


@pytest.mark.usefixtures('fake_source')
def test_ignore_structure_in_replace():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy tal:replace="structure _(u'føo')">Dummy</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'føo'


@pytest.mark.usefixtures('fake_source')
def test_repeat_multiple_assignment():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy tal:repeat="(ix, item) [(1, _(u'føo'))]">Dummy</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'føo'


@pytest.mark.usefixtures('fake_source')
def test_carriage_return_in_define():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy tal:define="foo True or
                                     _(u'føo')">Dummy</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'føo'


@pytest.mark.usefixtures('fake_source')
def test_multiline_replace():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy tal:replace="True or
                                      _(u'føo')">Dummy</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'føo'


@pytest.mark.usefixtures('fake_source')
def test_multiline_replace_with_structure():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <dummy tal:replace="structure True or
                                      _(u'føo')">Dummy</dummy>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'føo'


@pytest.mark.usefixtures('fake_source')
def test_spaces_around_tal_pipe_symbol():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <div tal:repeat="choice values | field.widget.values"/>
                </html>
                '''.encode('utf-8')
    list(xml_extractor('filename', _options()))


@pytest.mark.usefixtures('fake_source')
def test_empty_element():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <img title="message" i18n:attributes="title">
                  <p></p>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1


@pytest.mark.usefixtures('fake_source')
def test_brace_in_python_expression():
    global source
    source = u'''\
                <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <p>${some_method(_('abc'), {'a':'b'})}</p>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert len(messages) == 1
    assert messages[0].msgid == u'abc'


@pytest.mark.usefixtures('fake_source')
def test_translation_context():
    global source
    source = u'''<html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <button i18n:context="form" i18n:translate="">Save</button>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'Save'
    assert messages[0].msgctxt == u'form'


@pytest.mark.usefixtures('fake_source')
def test_translation_comment():
    global source
    source = u'''<html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <button i18n:comment="Generic save button" i18n:translate="">Save</button>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'Save'
    assert messages[0].comment == u'Generic save button'


@pytest.mark.usefixtures('fake_source')
def test_translation_comment_and_msgid():
    global source
    source = u'''<html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <button i18n:comment="Generic save button" i18n:translate="btn_save">Save</button>
                </html>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    assert messages[0].msgid == u'btn_save'
    assert messages[0].comment == u'Generic save button\nDefault: Save'


@pytest.mark.usefixtures('fake_source')
def test_linenumbers():
    global source
    source = u'''<!DOCTYPE html>
                 <html xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                      i18n:domain="lingua">
                  <span />
                  <span
                    />
                  <div>
                    <dummy i18n:comment="Generic save button" i18n:translate="dummy1">Foo<</dummy>
                  </div>
                  <div
                    foo="bar">
                    <dummy i18n:comment="Generic save button"
                           i18n:translate="dummy2">Foo</dummy>
                  </div>
                  <div foo="bar"
                    class="blubb">
                    <dummy i18n:comment="Generic
                                         save button"><span tal:omit-tag="" i18n:translate="dummy3">Foo</span></dummy>
                  </div>
                  <script>
                  // <![CDATA[

                    $(document).ready(function () {
                    });
                  // ]]>
                  </script>
                </html>

                <!--
                    comment
                -->
                <tal:macros
                    xmlns:i18n="http://xml.zope.org/namespaces/i18n"
                    i18n:domain="lingua">
                        <!--
                            comment
                        -->
                        <metal:action define-macro="action">
                            <dummy i18n:comment="Generic
                                                 save button">
                                <span tal:omit-tag="" i18n:translate="dummy4">Foo</span> ${value}
                            </dummy>
                        </metal:action>
                </tal:macros>
                '''.encode('utf-8')
    messages = list(xml_extractor('filename', _options()))
    got = [(x.msgid, x.location[1]) for x in messages]
    assert got == [
        ('dummy1', 8),
        ('dummy2', 13),
        ('dummy3', 18),
        ('dummy4', 41)]
